// -- [usi_twi_slave.c ] -----------------------------------------------------
// 	02.09.08
// 	AppNote      	: AVR312 - Using the USI module as a TWI slave
// 	Description    : USI-Slave simuliert PCF8574a
//							  TWI-Adresse kann per General_Call geändert werden.
//							  Es können mehrer Bytes empfangen werden,
//                     das Ende der Übertragung wird erkannt und per Flag signalisiert 
//							  Das Modul prüft USI_STOP und USI_START-Conditions.
//		Hardware			: ATtiny2313
//                     bei anderen mc's die SDA/SCL Pins überprüfen
//                     es werden GPIO Register benutzt - 
//                     falls die nicht existieren, auf volatile SRAM-Variablen ausweichen
//		Tabwidth			: 3
//
// ----------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include "usi_twi_slave.h"

static   uint8_t TWI_RxBuf[TWI_RX_BUFFER_SIZE];
static   uint8_t TWI_TxBuf[TWI_TX_BUFFER_SIZE];
volatile uint8_t TWI_SlaveAdress;

extern volatile uint8_t General_Call;

//-----------------------------------------------------------------------------
// Initialise USI for TWI Slave mode. Read TWI_Adress from Eeprom or use default Adress
void usi_twi_slave_init()
{
register uint8_t j;
																// Slave-Adresse ermitteln, dazu erstes Byte im Eeprom lesen 
j = eeprom_read_byte(0);								
if (j == 255) j = TWI_SLAVE_ADR;						//	Adresse aus Eeprom lesen oder Default-Wert übernehmen
j &= 0xFE;													// Bit.0 entfernen

TWI_SlaveAdress = j;

PORT_USI |=	(1<<PORT_USI_SCL | 1<<PORT_USI_SDA);// Set SCL high, SDA high
DDR_USI	|=	(1<<PORT_USI_SCL);						// Set SCL as output
DDR_USI &= ~(1<<PORT_USI_SDA);						// Set SDA as input
USICR 	=	(1<<USISIE)|(0<<USIOIE)|				// Enable Start Condition Interrupt. Disable Overflow Interrupt.
				(1<<USIWM1)|(0<<USIWM0)| 				// Set USI in Two-wire mode. No USI Counter overflow prior
				(1<<USICS1)|(0<<USICS0)|				// to first Start Condition (potentail failure)
				(0<<USICLK)|(0<<USITC);			  		// Shift Register Clock Source = External, positive edge
USISR = 0xF0;												// Clear all flags and reset overflow counter
USI_State = USI_WAIT;
}

//-----------------------------------------------------------------------------
// schreibt *msg in den Sendepuffer
void usi_twi_put_data( uint8_t *msg ,uint8_t msgSize )
{
uint8_t i;
// den Parameter mit der Arraygroesse vergleichen
if (msgSize > TWI_TX_BUFFER_SIZE) msgSize = TWI_TX_BUFFER_SIZE;

for ( i=0; i < msgSize; i++ )
TWI_TxBuf[i] = msg[i];
}

//-----------------------------------------------------------------------------
// liefert die Anzahl der empfangenen Bytes
uint8_t usi_twi_rx_bytes(void)
{
return(TWI_RxHead);
}

//-----------------------------------------------------------------------------
// kopiert Daten aus dem Empfangspuffer nach *msg
void usi_twi_get_data( uint8_t *msg ,uint8_t msgSize )
{
uint8_t i;
// den Parameter mit der Arraygroesse vergleichen
if ( msgSize > TWI_RX_BUFFER_SIZE)  msgSize = TWI_RX_BUFFER_SIZE;

for ( i = 0; i <  msgSize; i++ )
msg[i] = TWI_RxBuf[i];
TWI_RxHead = 0;										// nur einmal lesen - nicht immer nur nützlich !
}

// ----------------------------------------------------------------------------
// kann nach General Call aufgerufen werden -> ändert die TWI-Adresse im Eeprom
// dazu muss Parameter 1 = aktuelle Adresse sein
//           Parameter 2 = neue Adresse
//          (Parameter 3 = Konstante 0x55)
void eeprom_write_twi_adr(void)
{
															// Prüfen, bei Fehler zurückspringen
if (!(TWI_RxBuf[0] == TWI_SlaveAdress)) return;
//if (!(TWI_TxBuf[2] == 0x55)) return;			// dritter Parameter 0x55 zur Sicherheit

eeprom_write_byte(0, (TWI_RxBuf[1] & 0xFE));	// Bit.0 löschen und ins Eeprom schreiben

usi_twi_slave_init();								// Adresse übernehmen
}

//-----------------------------------------------------------------------------
//  Detects the USI_TWI Start Condition, waits for SCL to go low and then intialises the USI
//  for reception of the "TWI Address" packet (USI_State = USI_ADR).
ISR(USI_START_vect)
{															// Set default starting conditions for new TWI package
DDR_USI &= ~(1<<PORT_USI_SDA);					// Set SDA as input

while (PIN_USI & (1<<PORT_USI_SCL));			// Wait for SCL to go low
															// to ensure the "Start Condition" has completed.
USICR=(1<<USISIE)|(1<<USIOIE)|					// Enable Overflow and Start Condition Interrupt.
		(1<<USIWM1)|(1<<USIWM0)|					// (Keep StartCondInt to detect RESTART)
		(1<<USICS1)|(0<<USICS0)| 					// Set USI in Two-wire mode.
		(0<<USICLK)|(0<<USITC); 					// Shift Register Clock Source = External, positive edge

USISR=(1<<USISIF)|(1<<USIOIF)|					// Clear all flags
		(1<<USIPF) |(1<<USIDC) | 					
		(0x0<<USICNT0);								// Set USI to sample 8 bits i.e. count 16 external pin toggles.
USI_State = USI_ADR;									// Expect TWI address
}

//-----------------------------------------------------------------------------
// Handels all the comunication. Is disabled only when waiting for new Start Condition.
// USI_State zeigt auf den aktuellen Zustand des Moduls
ISR(USI_OVF_vect)
{
if (USI_State == USI_ADR)  						// --------------------- USI_ADR
	{
															// --------------------- Adressed by Master
	if ((( USIDR & 0xFE) == TWI_SlaveAdress) || USIDR == 0)
		{													// Pointer reset for receive/transmit data
		if (USIDR == 0) General_Call = TRUE;	// General Call recognized

		if (USIDR & 0x01)								// test Bit.0
			{
			TWI_TxTail = 0;
			USI_State = USI_WRITE;					// Write Data to Master
			}
		else
			{
			TWI_RxHead = 0;
			USI_State = USI_RECEIVE;				// Receive Data from Master
			}

		USIDR=0; 										// --------------------- send ACK

		DDR_USI |= (1<<PORT_USI_SDA);				// Set SDA as output
		USISR=(0<<USISIF)|(1<<USIOIF)|			// Clear all flags, except Start Cond
				(1<<USIPF) |(1<<USIDC) |
				(0xE<<USICNT0);						// Set USI to shift out 1 bit
															// Wait until ACK sent
															
		while ( !(USISR & ((1<<USISIF)|(1<<USIOIF)|(1<<USIPF )) ) ); // wait for a flag
		if (USISR & (1<<USISIF)) return;	      // wenn START-Flag gesetzt, dann beenden.
															// den Rest besorgt die USI-Start-ISR
															// das STOP-FLag wird ggf. weiter unten erkannt

		// das 1. Byte mit ACK ist durch, abhängig von USI_WRITE oder USI_RECEIVE gehts unterschiedlich weiter
		// und zwar mit der Vorbereitung zum Senden /Empfangen der nächsten 8 Bit

		if (USI_State == USI_WRITE)				// --------------------- write Byte
			{ // Byte in UDR schreiben und senden
			USIDR= TWI_TxBuf[TWI_TxTail++]; 		// Put data in USI Data Register.
															// Überlauf verhindern, TWI_TxTail ist Index
			if (TWI_TxTail > (TWI_TX_BUFFER_SIZE - 1)) TWI_TxTail = (TWI_TX_BUFFER_SIZE - 1) ;
			DDR_USI |= (1<<PORT_USI_SDA); 		// Set SDA as output

			// nachfolgend der Test auf USI-START oder USI-STOP während der ersten 2 Flanken
			USISR=(0<<USISIF)|(1<<USIOIF)|		// Clear all flags, except Start Cond
					(0<<USIPF) |(1<<USIDC) |
					(0xE<<USICNT0); 					// Set USI to sample 1 bit
															// Wait until 1 bit sampled or STOP condition
			while ( !(USISR & ((1<<USISIF)|(1<<USIOIF)|(1<<USIPF)) ) );
			
			if (USISR & (1<<USISIF)) return;		// --------------------- START_Flag		
															// wenn das START-Flag gesetzt, dann beenden.
															// den Rest besorgt die USI-Start-ISR

			if ( USISR & (1<<USIPF) )				// --------------------- STOP_Flag
			
				{											// Set USI in TWI Start Condition Mode
				USICR=(1<<USISIE)|(0<<USIOIE)|	// Enable Start Condition Interrupt. Disable Overflow Interrupt.
						(1<<USIWM1)|(0<<USIWM0)|	// Set USI in Two-wire mode. No USI Counter overflow hold.
						(1<<USICS1)|(0<<USICS0)|	// Shift Register Clock Source = External, positive edge
						(0<<USICLK)|(0<<USITC);
				USISR=(0<<USISIF)|(1<<USIOIF)		// Clear all flags but Start Condition
						|(1<<USIPF)|(1<<USIDC) |
						(0x0<<USICNT0);				// set USI to sample 8 bit
				USI_State = USI_WAIT;
				return;
				}

			else											// --------------------- USIOVF_Flag
				{
				USISR=(0<<USISIF)|(1<<USIOIF)|	// Clear all flags, except Start Cond
						(0<<USIPF) |(1<<USIDC) |
						(0x02<<USICNT0);				// Set USI to sample last 7 bits
				return;
				}
			}

		else												// --------------------- RECEIVE Byte
			{ 												// vorbereiten zum Empfang von 8 Bit
			DDR_USI &=~(1<<PORT_USI_SDA);			// Set SDA as input	**** überflüssig

			// nachfolgend der Test auf USI-START oder USI-STOP
			USISR=(0<<USISIF)|(1<<USIOIF)|		// Clear all flags, except Start Cond
					(0<<USIPF) |(1<<USIDC) |
					(0xE<<USICNT0); 					// Set USI to sample 1 bit
															// Wait until 1 bit sampled or STOP condition
			while ( !(USISR & ( (1<<USISIF)|(1<<USIOIF)|(1<<USIPF) )) );
			if (USISR & (1<<USISIF)) return;		// --------------------- START_FLAG	
															// wenn das START-Flag gesetzt, dann beenden.
															// den Rest besorgt die USI-Start-ISR

			if ( USISR & (1<<USIPF) )				// --------------------- STOP_Flag
				{
															// Set USI in TWI Start Condition Mode
				USICR=(1<<USISIE)|(0<<USIOIE)|	// Enable Start Condition Interrupt. Disable Overflow Interrupt.
						(1<<USIWM1)|(0<<USIWM0)|	// Set USI in Two-wire mode. No USI Counter overflow hold.
						(1<<USICS1)|(0<<USICS0)|	// Shift Register Clock Source = External, positive edge
						(0<<USICLK)|(0<<USITC);
				USISR=(0<<USISIF)|(1<<USIOIF)|	// Clear all flags but Start Condition
						(1<<USIPF) |(1<<USIDC) |
						(0x0<<USICNT0);				// Set USI to sample last 8 bits
				USI_State = USI_WAIT;
				return;
				}

			else											// --------------------- USIOVF_Flag
				{
				USISR=(0<<USISIF)|(1<<USIOIF)|	// Clear all flags, except Start Cond
						(0<<USIPF) |(1<<USIDC) |	
						(0x2<<USICNT0);				// Set USI to sample last 7 bits
				return;
				}
			}
		}
		
	else													// --------------------- NOT BEEN ADDRESSED BY MASTER
		{													// Set USI in TWI Start Condition Mode
		USICR=(1<<USISIE)|(0<<USIOIE)|			// Enable Start Condition Interrupt. Disable Overflow Interrupt.
				(1<<USIWM1)|(0<<USIWM0)|			// Set USI in Two-wire mode. No USI Counter overflow hold.
				(1<<USICS1)|(0<<USICS0)|			// Shift Register Clock Source = External, positive edge
				(0<<USICLK)|(0<<USITC);
		USISR=(0<<USISIF)|(1<<USIOIF)|			// Clear all flags but Start Condition
				(1<<USIPF) |(1<<USIDC) |
				(0x0<<USICNT0);						// Set USI to sample last 8 bits
		USI_State = USI_WAIT;
		return;
		}
	}

else if (USI_State == USI_WRITE)					// --------------------- WRITE 
	{ 														// ein Byte ist geschrieben worden
															// --------------------- WAIT FOR ACK/NACK
	DDR_USI &=~(1<<PORT_USI_SDA);					// Set SDA as input
	USIDR=0;
	USISR=(0<<USISIF)|(1<<USIOIF)| 				// Clear all flags, except Start Cond
			(1<<USIPF) |(1<<USIDC)|					// Set USI to sample 1 bit
			(0xE<<USICNT0);							// Wait until (N)ACK received. (Pass if Start Cond.)
			
	while ( !(USISR & ((1<<USISIF)|(1<<USIOIF)|(1<<USIPF)) ) );
	if (USISR & (1<<USISIF)) return;				// --------------------- START_Flag
															// Beenden, den Rest besorgt die USI-Start-ISR

	if ( USIDR )										// --------------------- NACK received
		{
		USICR=(1<<USISIE)|(0<<USIOIE)| 			// Enable Start Condition Interrupt. Disable Overflow Interrupt.
				(1<<USIWM1)|(0<<USIWM0)| 			// Set USI in Two-wire mode. No USI Counter overflow hold.
				(1<<USICS1)|(0<<USICS0)| 			// Shift Register Clock Source = External, positive edge
				(0<<USICLK)|(0<<USITC);
		USISR=(0<<USISIF)|(1<<USIOIF)|			// Clear all flags but Start Condition
				(1<<USIPF) |(1<<USIDC) |
				(0x0<<USICNT0);						// Set USI to sample last 8 bits
		USI_State = USI_NACK;
		return;
		}
															// --------------------- ACK received
															// das nächste Byte zum Senden vorbereiten
	if (TWI_TxTail < (TWI_TX_BUFFER_SIZE - 1))// Überlauf verhindern
		{
		USIDR= TWI_TxBuf[TWI_TxTail++]; 			// Put data in USI Data Register and increment Index
		}

	else													// Index nicht weiter incrementieren
		{
		USIDR = TWI_TxBuf[TWI_TxTail];			// Put data in USI Data Register.
		}
	DDR_USI |= (1<<PORT_USI_SDA); 				// Set SDA as output

															// nachfolgend der Test auf USI-START oder USI-STOP
	USISR=(0<<USISIF)|(1<<USIOIF)|				// Clear all flags, except Start Cond
			(1<<USIPF) |(1<<USIDC) | 				// Set USI to sample 1 bit
			(0xE<<USICNT0);							// Wait until 1 bit sampled or STOP condition
	
	while ( !(USISR & ((1<<USISIF)|(1<<USIOIF)|(1<<USIPF)) ) );
	
	if (USISR & (1<<USISIF)) return;				// --------------------- START_Flag
															// Beenden, den Rest besorgt die USI-Start-ISR
	else if ( USISR & (1<<USIPF) )				// --------------------- STOP_Flag
		{
															// Set USI in TWI Start Condition Mode
		USICR=(1<<USISIE)|(0<<USIOIE)|			// Enable Start Condition Interrupt. Disable Overflow Interrupt.
				(1<<USIWM1)|(0<<USIWM0)|			// Set USI in Two-wire mode. No USI Counter overflow hold.
				(1<<USICS1)|(0<<USICS0)|			// Shift Register Clock Source = External, positive edge
				(0<<USICLK)|(0<<USITC);
		USISR=(0<<USISIF)|(1<<USIOIF)|			// Clear all flags but Start Condition
				(1<<USIPF) |(1<<USIDC) |
				(0x0<<USICNT0);						// Set USI to sample last 8 bits
		USI_State = USI_STOP;
		return;
		}
															// --------------------- USIOVF_Flag
	USISR=(0<<USISIF)|(1<<USIOIF)|				// Clear all flags, except Start Cond
			(1<<USIPF) |(1<<USIDC) |			
			(0x2<<USICNT0);							// Set USI to sample last 7 bits
	return;
	}

else if (USI_State == USI_RECEIVE) 				// --------------------- RECEIVE
	{														// --------------------- Byte speichern
	if (TWI_RxHead < (TWI_RX_BUFFER_SIZE)) // nur speichern, wenn innerhalb des gültigen Index
		{													// kommen mehr Bytes an, dann werden sie ignoriert
		TWI_RxBuf[TWI_RxHead++] = USIDR;			// TWI_RxHead = Index + 1 !! = Anzahl der gelesen Bytes	
		}
	USIDR=0;		 										// ------------------- Send ACK

	DDR_USI |=(1<<PORT_USI_SDA);					// Set SDA as output
	USISR=(0<<USISIF)|(1<<USIOIF)|				// Clear all flags, except Start Cond
			(1<<USIPF) |(1<<USIDC) |				
			(0xE<<USICNT0);							// Set USI to shift out 1 bit												

	while ( !(USISR & ((1<<USISIF)|(1<<USIOIF)|(1<<USIPF)) ) );
	
	if (USISR & (1<<USISIF)) return;				// --------------------- START_Flag
															// Beenden, den Rest besorgt die USI-Start-ISR
															
	else if ( USISR & (1<<USIPF) )				// --------------------- STOP_Flag
		{													// Set USI in TWI Start Condition Mode
		USICR=(1<<USISIE)|(0<<USIOIE)|			// Enable Start Condition Interrupt. Disable Overflow Interrupt.
				(1<<USIWM1)|(0<<USIWM0)|			// Set USI in Two-wire mode. No USI Counter overflow hold.
				(1<<USICS1)|(0<<USICS0)|			// Shift Register Clock Source = External, positive edge
				(0<<USICLK)|(0<<USITC);
		USISR=(0<<USISIF)|(1<<USIOIF)|			// Clear all flags but Start Condition
				(1<<USIPF) |(1<<USIDC) |
				(0x0<<USICNT0);						// Set USI to shift out 8 bit															
		USI_State = USI_STOP;
		return;															
		}													
															// --------------------- prepare to receive next Byte
	DDR_USI &=~(1<<PORT_USI_SDA);					// Set SDA as input
															// nachfolgend der Test auf USI-START oder USI-STOP
	USISR=(0<<USISIF)|(1<<USIOIF)|				// Clear all flags, except Start Cond
			(1<<USIPF) |(1<<USIDC) |
			(0xE<<USICNT0); 							// Set USI to sample 1 bit
															// Wait until 1 bit sampled or STOP/START condition
	while ( !(USISR & ((1<<USISIF)|(1<<USIOIF)|(1<<USIPF)) ) );
	
	if (USISR & (1<<USISIF)) return;				// --------------------- START_Flag
															// den Rest besorgt die USI-Start-ISR
	else if ( USISR & (1<<USIPF) )				// --------------------- STOP_Flag
		{													// Set USI in TWI Start Condition Mode
		USICR=(1<<USISIE)|(0<<USIOIE)|			// Enable Start Condition Interrupt. Disable Overflow Interrupt.
				(1<<USIWM1)|(0<<USIWM0)|			// Set USI in Two-wire mode. No USI Counter overflow hold.
				(1<<USICS1)|(0<<USICS0)|			// Shift Register Clock Source = External, positive edge
				(0<<USICLK)|(0<<USITC);
		USISR=(0<<USISIF)|(1<<USIOIF)|			// Clear all flags but Start Condition
				(1<<USIPF) |(1<<USIDC) |
				(0x0<<USICNT0);						// Set USI to shift out 8 bit
		USI_State = USI_STOP;
		return;
		}
															// --------------------- USIOVF_Flag
	USISR=(0<<USISIF)|(1<<USIOIF)|				// Clear all flags, except Start Cond
			(1<<USIPF) |(1<<USIDC) |
			(0x2<<USICNT0);							// Set USI to sample last 7 bits
	return;
	}
}

